@@ -458,7 +458,7 @@ Debug information is inserted in the buffer \"*POP3 DEBUG*\"")
 	      (setq session (elmo-pop3-get-session folder))
 	    (if session
 		(elmo-network-close-session session)))))
-    t))
+    (file-directory-p (elmo-folder-msgdb-path folder))))
 
 (defun elmo-pop3-parse-uidl-response (string)
   (let ((buffer (current-buffer))
@@ -576,7 +576,7 @@ Debug information is inserted in the buffer \"*POP3 DEBUG*\"")
       (elmo-pop3-list-by-uidl-subr folder)
     (elmo-pop3-list-by-list folder)))
 
-(luna-define-method elmo-folder-list-messages-internal
+(luna-define-method elmo-folder-list-messages-plugged
   ((folder elmo-pop3-folder) &optional nohide)
   (elmo-pop3-folder-list-messages folder))
 
