@@ -200,7 +200,7 @@ Remove Re, Was, Fwd etc."
 
 (defun wl-score-string-index< (a1 a2)
   (string-lessp (wl-score-ov-entity-get-by-index (car a1) wl-score-index)
- 		(wl-score-ov-entity-get-by-index (car a2) wl-score-index)))
+		(wl-score-ov-entity-get-by-index (car a2) wl-score-index)))
 
 (defun wl-score-string-func< (a1 a2)
   (string-lessp (funcall wl-score-index (car a1))
@@ -341,7 +341,7 @@ Set `wl-score-cache' nil."
   (let ((files (wl-get-assoc-list-value
 		score-alist folder
 		(if (not wl-score-folder-alist-matchone) 'all-list)))
-        fl f)
+	fl f)
     (while (setq f (wl-pop files))
       (wl-append
        fl
@@ -1206,6 +1206,7 @@ Set `wl-score-cache' nil."
 	 (count (length alist))
 	 (i 0)
 	 (update-unread nil)
+	 wl-summary-unread-message-hook
 	 num score dels visible score-mark mark-alist)
     (save-excursion
       (message "Updating score...")
@@ -1224,9 +1225,9 @@ Set `wl-score-cache' nil."
 	       (wl-push num dels))
 	      ((< score wl-summary-mark-below)
 	       (if visible
-		   (wl-summary-mark-as-read t); opened
+		   (wl-summary-mark-as-read num); opened
 		 (setq update-unread t)
-		 (wl-summary-mark-as-read t nil nil num))) ; closed
+		 (wl-summary-mark-as-read num))) ; closed
 	      ((and wl-summary-important-above
 		    (> score wl-summary-important-above))
 	       (if (wl-thread-jump-to-msg num);; force open
@@ -1244,15 +1245,12 @@ Set `wl-score-cache' nil."
 	   'wl-summary-score-update-all-lines "Updating score..."
 	   (/ (* i 100) count))))
       (when dels
-	(setq mark-alist
-	      (elmo-msgdb-get-mark-alist (wl-summary-buffer-msgdb)))
 	(let ((marks dels))
 	  (while marks
-	    (setq mark-alist
-		  (elmo-msgdb-mark-set mark-alist (pop marks) nil))))
+	    (elmo-msgdb-set-mark (wl-summary-buffer-msgdb)
+				 (pop marks) nil)))
 	(elmo-folder-mark-as-read wl-summary-buffer-elmo-folder
 				  dels)
-	(elmo-msgdb-set-mark-alist (wl-summary-buffer-msgdb) mark-alist)
 	(wl-summary-delete-messages-on-buffer dels))
       (when (and update update-unread)
 	(let ((num-db (elmo-msgdb-get-number-alist
@@ -1261,10 +1259,12 @@ Set `wl-score-cache' nil."
 			   (wl-summary-buffer-msgdb))))
 	  ;; Update Folder mode
 	  (wl-folder-set-folder-updated (wl-summary-buffer-folder-name)
-					(list 0
-					      (wl-summary-count-unread
-					       mark-alist)
-					      (length num-db)))
+					(list 
+					 0
+					 (let ((pair
+						(wl-summary-count-unread)))
+					   (+ (car pair) (cdr pair)))
+					 (length num-db)))
 	  (wl-summary-update-modeline)))
       (message "Updating score...done")
       dels)))
